@@ -6,9 +6,11 @@ import dateutil.tz
 import pytz
 import datetime
 import xml.dom.minidom
+from sql import Table, Column
+
 from trytond.model import ModelSQL, ModelView, fields
 from trytond.tools import reduce_ids
-from trytond.backend import TableHandler
+from trytond import backend
 from trytond.pyson import Eval, If, Bool
 from trytond.transaction import Transaction
 from trytond.pool import Pool
@@ -80,7 +82,7 @@ class Todo(ModelSQL, ModelView):
         ('completed', 'Completed'),
         ('in-process', 'In-Process'),
         ('cancelled', 'Cancelled'),
-        ], 'Status', on_change=['status', 'completed', 'percent_complete'])
+        ], 'Status')
     summary = fields.Char('Summary')
     uuid = fields.Char('UUID', required=True,
             help='Universally Unique Identifier', select=True)
@@ -153,16 +155,11 @@ class Todo(ModelSQL, ModelView):
     def default_classification():
         return 'public'
 
-    @staticmethod
-    def default_timezone():
-        User = Pool().get('res.user')
-        user = User(Transaction().user)
-        return user.timezone
-
     @staticmethod
     def default_percent_complete():
         return 0
 
+    @fields.depends('status', 'completed', 'percent_complete')
     def on_change_status(self):
         res = {}
         if not getattr(self, 'status', None):
@@ -191,7 +188,7 @@ class Todo(ModelSQL, ModelView):
 
     @classmethod
     def validate(cls, todos):
-        super(todos, cls).validate(todos)
+        super(Todo, cls).validate(todos)
         for todo in todos:
             todo.check_recurrence()
 
@@ -275,105 +272,115 @@ class Todo(ModelSQL, ModelView):
         res['location'] = self.location.id
         res['status'] = self.status
         res['organizer'] = self.organizer
-        res['rdates'] = [('delete_all',)]
+        res['rdates'] = [('delete', [r.id for r in self.rdates])]
         to_create = [rdate._date2update() for rdate in self.rdates]
         if to_create:
             res['rdates'].append(('create', to_create))
-        res['exdates'] = [('delete_all',)]
+        res['exdates'] = [('delete', [r.id for r in self.exdates])]
         to_create = [exdate._date2update() for exdate in self.exdates]
         if to_create:
             res['exdates'].append(('create', to_create))
-        res['rrules'] = [('delete_all',)]
+        res['rrules'] = [('delete', [r.id for r in self.rrules])]
         to_create = [rrule._rule2update() for rrule in self.rrules]
         if to_create:
             res['rrules'].append(('create', to_create))
-        res['exrules'] = [('delete_all',)]
+        res['exrules'] = [('delete', [r.id for r in self.exrules])]
         to_create = [exrule._rule2update() for exrule in self.exrules]
         if to_create:
             res['exrules'].append(('create', to_create))
         return res
 
     @classmethod
-    def write(cls, todos, values):
+    def write(cls, *args):
         pool = Pool()
         Calendar = pool.get('calendar.calendar')
         Collection = pool.get('webdav.collection')
+        table = cls.__table__()
 
         cursor = Transaction().cursor
 
-        values = values.copy()
-        if 'sequence' in values:
-            del values['sequence']
+        actions = iter(args)
+        args = []
+        for todos, values in zip(actions, actions):
+            values = values.copy()
+            if 'sequence' in values:
+                del values['sequence']
+            args.extend((todos, values))
 
-        super(Todo, cls).write(todos, values)
+        super(Todo, cls).write(*args)
 
         ids = [t.id for t in todos]
         for i in range(0, len(ids), cursor.IN_MAX):
             sub_ids = ids[i:i + cursor.IN_MAX]
-            red_sql, red_ids = reduce_ids('id', sub_ids)
-            cursor.execute('UPDATE "' + cls._table + '" '
-                'SET sequence = sequence + 1 '
-                'WHERE ' + red_sql, red_ids)
+            red_sql = reduce_ids(table.id, sub_ids)
+            cursor.execute(*table.update(
+                    columns=[table.sequence],
+                    values=[table.sequence + 1],
+                    where=red_sql))
 
-        for todo in todos:
-            if (todo.calendar.owner
-                    and (todo.organizer == todo.calendar.owner.email
-                        or (todo.parent
-                            and todo.parent.organizer
-                            == todo.calendar.owner.email))):
-                if todo.organizer == todo.calendar.owner.email:
-                    attendee_emails = [x.email for x in todo.attendees
-                            if x.status != 'declined'
-                            and x.email != todo.organizer]
-                else:
-                    attendee_emails = [x.email for x in todo.parent.attendees
-                            if x.status != 'declined'
-                            and x.email != todo.parent.organizer]
-                if attendee_emails:
-                    with Transaction().set_user(0):
-                        todo2s = cls.search([
-                            ('uuid', '=', todo.uuid),
-                            ('calendar.owner.email', 'in', attendee_emails),
-                            ('id', '!=', todo.id),
-                            ('recurrence', '=', todo.recurrence),
-                            ])
-                    for todo2 in todo2s:
-                        if todo2.calendar.owner.email in attendee_emails:
-                            attendee_emails.remove(todo2.calendar.owner.email)
-                    with Transaction().set_user(0):
-                        cls.write(todos, todo._todo2update())
-                if attendee_emails:
-                    with Transaction().set_user(0):
-                        calendars = Calendar.search([
-                            ('owner.email', 'in', attendee_emails),
-                            ])
-                        if not todo.recurrence:
-                            for calendar in calendars:
-                                new_todo, = cls.copy([todo], default={
-                                    'calendar': calendar.id,
-                                    'occurences': None,
-                                    'uuid': todo.uuid,
-                                    })
-                                for occurence in todo.occurences:
-                                    cls.copy([occurence], default={
+        actions = iter(args)
+        for todos, values in zip(actions, actions):
+            if not values:
+                continue
+            for todo in todos:
+                if (todo.calendar.owner
+                        and (todo.organizer == todo.calendar.owner.email
+                            or (todo.parent
+                                and todo.parent.organizer
+                                == todo.calendar.owner.email))):
+                    if todo.organizer == todo.calendar.owner.email:
+                        attendee_emails = [x.email for x in todo.attendees
+                                if x.status != 'declined'
+                                and x.email != todo.organizer]
+                    else:
+                        attendee_emails = [x.email for x in todo.parent.attendees
+                                if x.status != 'declined'
+                                and x.email != todo.parent.organizer]
+                    if attendee_emails:
+                        with Transaction().set_user(0):
+                            todo2s = cls.search([
+                                ('uuid', '=', todo.uuid),
+                                ('calendar.owner.email', 'in', attendee_emails),
+                                ('id', '!=', todo.id),
+                                ('recurrence', '=', todo.recurrence),
+                                ])
+                        for todo2 in todo2s:
+                            if todo2.calendar.owner.email in attendee_emails:
+                                attendee_emails.remove(todo2.calendar.owner.email)
+                        with Transaction().set_user(0):
+                            cls.write(todos, todo._todo2update())
+                    if attendee_emails:
+                        with Transaction().set_user(0):
+                            calendars = Calendar.search([
+                                ('owner.email', 'in', attendee_emails),
+                                ])
+                            if not todo.recurrence:
+                                for calendar in calendars:
+                                    new_todo, = cls.copy([todo], default={
                                         'calendar': calendar.id,
-                                        'parent': new_todo.id,
-                                        'uuid': occurence.uuid,
+                                        'occurences': None,
+                                        'uuid': todo.uuid,
+                                        })
+                                    for occurence in todo.occurences:
+                                        cls.copy([occurence], default={
+                                            'calendar': calendar.id,
+                                            'parent': new_todo.id,
+                                            'uuid': occurence.uuid,
+                                            })
+                            else:
+                                parents = cls.search([
+                                        ('uuid', '=', todo.uuid),
+                                        ('calendar.owner.email', 'in',
+                                            attendee_emails),
+                                        ('id', '!=', todo.id),
+                                        ('recurrence', '=', None),
+                                        ])
+                                for parent in parents:
+                                    cls.copy([todo], default={
+                                        'calendar': parent.calendar.id,
+                                        'parent': parent.id,
+                                        'uuid': todo.uuid,
                                         })
-                        else:
-                            parents = cls.search([
-                                    ('uuid', '=', todo.uuid),
-                                    ('calendar.owner.email', 'in',
-                                        attendee_emails),
-                                    ('id', '!=', todo.id),
-                                    ('recurrence', '=', None),
-                                    ])
-                            for parent in parents:
-                                cls.copy([todo], default={
-                                    'calendar': parent.calendar.id,
-                                    'parent': parent.id,
-                                    'uuid': todo.uuid,
-                                    })
         # Restart the cache for todo
         Collection._todo_cache.clear()
 
@@ -489,7 +496,7 @@ class Todo(ModelSQL, ModelView):
         if hasattr(vtodo, 'percent_complete'):
             res['percent_complete'] = int(vtodo.percent_complete.value)
         else:
-            res['percent_complete'] = None
+            res['percent_complete'] = 0
 
         if hasattr(vtodo, 'completed'):
             if not isinstance(vtodo.completed.value, datetime.datetime):
@@ -538,6 +545,8 @@ class Todo(ModelSQL, ModelView):
             res['status'] = vtodo.status.value.lower()
         else:
             res['status'] = ''
+
+        res['categories'] = [('remove', [c.id for c in todo.categories])]
         if hasattr(vtodo, 'categories'):
             categories = Category.search([
                 ('name', 'in', [x for x in vtodo.categories.value]),
@@ -553,9 +562,7 @@ class Todo(ModelSQL, ModelView):
                             })
             if to_create:
                 categories += Category.create(to_create)
-            res['categories'] = [('set', [c.id for c in categories])]
-        else:
-            res['categories'] = [('unlink_all',)]
+            res['categories'] += [('add', [c.id for c in categories])]
         if hasattr(vtodo, 'class'):
             if getattr(vtodo, 'class').value.lower() in \
                     dict(cls.classification.selection):
@@ -708,14 +715,8 @@ class Todo(ModelSQL, ModelView):
         '''
         Return an iCalendar instance of vobject for todo
         '''
-        pool = Pool()
-        User = pool.get('res.user')
-
-        user = User(Transaction().user)
         if self.timezone:
             tztodo = dateutil.tz.gettz(self.timezone)
-        elif user.timezone:
-            tztodo = dateutil.tz.gettz(user.timezone)
         else:
             tztodo = tzlocal
 
@@ -809,7 +810,7 @@ class Todo(ModelSQL, ModelView):
             vtodo.add('class')
             getattr(vtodo, 'class').value = self.classification.upper()
         elif getattr(vtodo, 'class').value.lower() in \
-                dict(self.classification.selection):
+                dict(self.__class__.classification.selection):
             getattr(vtodo, 'class').value = self.classification.upper()
         if self.location:
             if not hasattr(vtodo, 'location'):
@@ -879,7 +880,9 @@ class TodoRDate(DateMixin, ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
+        TableHandler = backend.get('TableHandler')
         cursor = Transaction().cursor
+        sql_table = cls.__table__()
         # Migration from 1.4: calendar_rdate renamed to calendar_date
         table = TableHandler(cursor, cls, module_name)
         old_column = 'calendar_rdate'
@@ -892,13 +895,13 @@ class TodoRDate(DateMixin, ModelSQL, ModelView):
 
         # Migration from 2.6: Remove inherits calendar.date
         if table.column_exist('calendar_date'):
-            cursor.execute('UPDATE "' + cls._table + '" AS e '
-                'SET date = (SELECT a.date '
-                    'FROM calendar_date AS a '
-                    'WHERE a.id = e.calendar_date), '
-                'datetime = (SELECT a.datetime '
-                    'FROM calendar_date AS a '
-                    'WHERE a.id = e.calendar_date)')
+            date = Table('calendar_date')
+            cursor.execute(*sql_table.update(
+                    columns=[sql_table.date, sql_table.datetime],
+                    values=[date.select(date.date,
+                            where=date.id == sql_table.calendar_date),
+                        date.select(date.datetime,
+                            where=date.id == sql_table.calendar_date)]))
             table.drop_column('calendar_date', True)
 
     @classmethod
@@ -914,15 +917,19 @@ class TodoRDate(DateMixin, ModelSQL, ModelView):
         return super(TodoRDate, cls).create(vlist)
 
     @classmethod
-    def write(cls, rdates, values):
+    def write(cls, *args):
         Todo = Pool().get('calendar.todo')
-        todos = [x.todo for x in rdates]
-        if values.get('todo'):
-            todos.append(Todo(values['todo']))
+
+        actions = iter(args)
+        todos = []
+        for rdates, values in zip(actions, actions):
+            todos += [x.todo for x in rdates]
+            if values.get('todo'):
+                todos.append(Todo(values['todo']))
         if todos:
             # Update write_date of todo
             Todo.write(todos, {})
-        super(TodoRDate, cls).write(rdates, values)
+        super(TodoRDate, cls).write(*args)
 
     @classmethod
     def delete(cls, todo_rdates):
@@ -944,7 +951,9 @@ class TodoRRule(RRuleMixin, ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
+        TableHandler = backend.get('TableHandler')
         cursor = Transaction().cursor
+        sql_table = cls.__table__()
 
         super(TodoRRule, cls).__register__(module_name)
 
@@ -952,13 +961,13 @@ class TodoRRule(RRuleMixin, ModelSQL, ModelView):
 
         # Migration from 2.6: Remove inherits calendar.rrule
         if table.column_exist('calendar_rrule'):
+            rrule = Table('calendar_rrule')
             for field in (f for f in dir(RRuleMixin)
                     if isinstance(f, fields.Field)):
-                cursor.execute(('UPDATE "' + cls._table + '" AS e '
-                        'SET "%(field)s" = (SELECT a."%(field)s" '
-                            'FROM calendar_rrule AS r '
-                            'WHERE r.id = e.calendar_rrule)')
-                    % {'field': field})
+                cursor.execute(*sql_table.update(
+                        columns=[Column(sql_table, field)],
+                        values=[rrule.select(Column(rrule, field),
+                                where=rrule.id == sql_table.calendar_rrule)]))
             table.drop_column('calendar_rrule', True)
 
     @classmethod
@@ -974,15 +983,19 @@ class TodoRRule(RRuleMixin, ModelSQL, ModelView):
         return super(TodoRRule, cls).create(vlist)
 
     @classmethod
-    def write(cls, todo_rrules, values):
+    def write(cls, *args):
         Todo = Pool().get('calendar.todo')
-        todos = [x.todo for x in todo_rrules]
-        if values.get('todo'):
-            todos.append(Todo(values['todo']))
+
+        actions = iter(args)
+        todos = []
+        for todo_rrules, values in zip(actions, actions):
+            todos += [x.todo for x in todo_rrules]
+            if values.get('todo'):
+                todos.append(Todo(values['todo']))
         if todos:
             # Update write_date of todo
             Todo.write(todos, {})
-        super(TodoRRule, cls).write(todo_rrules, values)
+        super(TodoRRule, cls).write(*args)
 
     @classmethod
     def delete(cls, todo_rrules):
@@ -1015,7 +1028,9 @@ class TodoAttendee(AttendeeMixin, ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
+        TableHandler = backend.get('TableHandler')
         cursor = Transaction().cursor
+        sql_table = cls.__table__()
 
         super(TodoAttendee, cls).__register__(module_name)
 
@@ -1023,13 +1038,14 @@ class TodoAttendee(AttendeeMixin, ModelSQL, ModelView):
 
         # Migration from 2.6: Remove inherits calendar.attendee
         if table.column_exist('calendar_attendee'):
-            cursor.execute('UPDATE "' + cls._table + '" AS e '
-                'SET email = (SELECT a.email '
-                    'FROM calendar_attendee AS a '
-                    'WHERE a.id = e.calendar_attendee), '
-                'status = (SELECT a.status '
-                    'FROM calendar_attendee AS a '
-                    'WHERE a.id = e.calendar_attendee)')
+            attendee = Table('calendar_attendee')
+            cursor.execute(*sql_table.update(
+                    columns=[sql_table.email, sql_table.status],
+                    values=[attendee.select(attendee.email,
+                            where=attendee.id == sql_table.calendar_attendee),
+                        attendee.select(attendee.status,
+                            where=
+                            attendee.id == sql_table.calendar_attendee)]))
             table.drop_column('calendar_attendee', True)
 
     @classmethod
@@ -1072,23 +1088,29 @@ class TodoAttendee(AttendeeMixin, ModelSQL, ModelView):
         return attendees
 
     @classmethod
-    def write(cls, attendees, values):
+    def write(cls, *args):
         Todo = Pool().get('calendar.todo')
 
-        todos = [x.todo.id for x in attendees]
-        if values.get('todo'):
-            todos.append(Todo(values['todo']))
+        actions = iter(args)
+        args = []
+        todos = []
+        for todo_attendees, values in zip(actions, actions):
+            todos += [x.todo.id for x in todo_attendees]
+            if values.get('todo'):
+                todos.append(Todo(values['todo']))
+            if 'email' in values:
+                values = values.copy()
+                del values['email']
+            args.extend((todo_attendees, values))
+
         if todos:
             # Update write_date of todo
             Todo.write(todos, {})
 
-        if 'email' in values:
-            values = values.copy()
-            del values['email']
+        super(TodoAttendee, cls).write(*args)
 
-        super(TodoAttendee, cls).write(attendees, values)
-        for attendee in attendees:
-            todo = attendee.todo
+        for todo_attendee in sum(args[::2], []):
+            todo = todo_attendee.todo
             if (todo.calendar.owner
                     and (todo.organizer == todo.calendar.owner.email
                         or (todo.parent
@@ -1106,11 +1128,11 @@ class TodoAttendee(AttendeeMixin, ModelSQL, ModelView):
                                 ('todo.uuid', '=', todo.uuid),
                                 ('todo.calendar.owner.email', 'in',
                                     attendee_emails),
-                                ('id', '!=', attendee.id),
+                                ('id', '!=', todo_attendee.id),
                                 ('todo.recurrence', '=', todo.recurrence),
-                                ('email', '=', attendee.email),
+                                ('email', '=', todo_attendee.email),
                                 ])
-                        cls.write(attendees2, attendee._attendee2update())
+                        cls.write(attendees2, todo_attendee._attendee2update())
 
     @classmethod
     def delete(cls, todo_attendees):
@@ -1177,7 +1199,9 @@ class TodoAlarm(AlarmMixin, ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
+        TableHandler = backend.get('TableHandler')
         cursor = Transaction().cursor
+        sql_table = cls.__table__()
 
         super(TodoAlarm, cls).__register__(module_name)
 
@@ -1185,10 +1209,11 @@ class TodoAlarm(AlarmMixin, ModelSQL, ModelView):
 
         # Migration from 2.6: Remove inherits calendar.alarm
         if table.column_exist('calendar_alarm'):
-            cursor.execute('UPDATE "' + cls._table + '" AS t '
-                'SET valarm = (SELECT a.valarm '
-                    'FROM calendar_alarm AS a '
-                    'WHERE a.id = t.calendar_alarm)')
+            alarm = Table('calendar_alarm')
+            cursor.execute(*sql_table.update(
+                    columns=[sql_table.valarm],
+                    values=[alarm.select(alarm.valarm,
+                            where=alarm.id == sql_table.calendar_alarm)]))
             table.drop_column('calendar_alarm', True)
 
     @classmethod
@@ -1204,15 +1229,19 @@ class TodoAlarm(AlarmMixin, ModelSQL, ModelView):
         return super(TodoAlarm, cls).create(vlist)
 
     @classmethod
-    def write(cls, alarms, values):
+    def write(cls, *args):
         Todo = Pool().get('calendar.todo')
-        todos = [x.todo for x in alarms]
-        if values.get('todo'):
-            todos.append(Todo(values['todo']))
+
+        actions = iter(args)
+        todos = []
+        for alarms, values in zip(actions, actions):
+            todos += [x.todo for x in alarms]
+            if values.get('todo'):
+                todos.append(Todo(values['todo']))
         if todos:
             # Update write_date of todo
             Todo.write(todos, {})
-        super(TodoAlarm, cls).write(alarms, values)
+        super(TodoAlarm, cls).write(*args)
 
     @classmethod
     def delete(cls, todo_alarms):
